library IEEE;
use IEEE.std_logic_1164.all;

library work;
use work.settings.all;

-- Simulation of the comparators switching behavior
-- The LVDS and LVPECL input buffers can not be simulated. Thus, the OPAMP module generates combinatorical logic 
-- (VP and not VN) instead of a differential input buffer. This module translates two real valued signals (the original
-- comparator inputs to be compared against each other) into two logic signals to be passed to the OPAMP module.
-- 
-- author: Andreas Engel
-- date:   03/13
entity OPAMP_SIM is
  port (
    VPR : in  real;        -- real valued positive comparator input     
    VNR : in  real;        -- real valued negative comparator input
    VP  : out std_logic;   -- logic equivalent of positive comparator input
    VN  : out std_logic);  -- logic equivalent of negative comparator input
end entity;
------------------------------------------------------------------------------------------------------------------------

architecture beh of OPAMP_SIM is
  signal s_diff : real                          := 0.0;
  signal s_lfsr : std_logic_vector(17 downto 0) := (others => '0');
begin

  p_opamp: process
    variable v_diff : real;
    variable v_high : real;
    variable v_low  : real;
  begin
    wait until (VPR < 10.0 and VNR < 10.0);
  
    v_diff := (VPR - VNR) * 1000.0; -- diff in mV
    s_diff <= v_diff;
    
    -- edge detection for hysteresis simulation
    if (v_diff > s_diff) then
      v_high := OPAMP_VRH;
      v_low  := OPAMP_VRL;
    else
      v_high := OPAMP_VFH;
      v_low  := OPAMP_VFL;
    end if;
    
    -- instable comparator decision inside switching limits
    if    (v_diff > v_high) then VP <= '1';
    elsif (v_diff < v_low)  then VP <= '0';
    else                         VP <= s_lfsr(0);
    end if;
    
    -- pseudo random number generator
    s_lfsr <= s_lfsr(16 downto 0) & (s_lfsr(17) xnor s_lfsr(10));
    
    -- fixed update rate
    wait for OPAMP_UPDATE * 1.0 ns;
  end process;
  
  VN <= '0';
  
end architecture;
