`default_nettype none
`timescale 1ns/100ps

// Testbench for comparator project.
//  * generates 48 MHz OSC_CLK and reset Signals (PO_RSTN, PB_RSTN)
//  * generates comparator inputs:
//    * VP = 1.5V +- 0.5V ramp, 5 kHz
//    * VN = 1.5V dc
//  * simulates the comparators non-ideal switching behavior
//  * applies various clock divider settings to the SWITCH input
//  * checks period and duty cycle of CLK and DCLK
//  * checks synchronization of RSTN and DRSTN to the corresponding clock signals
//  * checks LED, LOCK and COMPARE output
// 
// autor: Andreas Engel
// date:  03/2013
module TB #(`include "settings.v");
`include "test.v"
/**********************************************************************************************************************/
  
// clock and reset generation
reg                   osc_clk;
initial               osc_clk = 0;
always  #(500.0/48.0) osc_clk = ~osc_clk;
reg                   po_rstN;
initial               po_rstN = 0;
initial #(300.0/48.0) po_rstN = 1;
reg                   pb_rstN;
initial               pb_rstN = 0;
initial #(200.0/48.0) pb_rstN = 1;
  
// simulate positive/negative comparator input
real vpr = 1.0;
real vnr = 1.5;
always #10 vpr <= vpr + 0.0001 * (($stime % 200000) < 100000 ? 1 : -1);

// simulate comparator behavoir
wire vp;
wire vn;
OPAMP_SIM i_opamp (
  .VPR($realtobits(vpr)),
  .VNR($realtobits(vnr)),
  .VP (vp),
  .VN (vn));

// each divider selected for 6 dclk cycles
// (disable period/duty cycle assertions for first rising edege after divider change)
wire lock;
wire dclk;
reg  dclk_assert;
integer i,div;
initial begin
  div         <= 1;
  dclk_assert <= 0;
  @(posedge lock);
  for (div = 1; div <= 16; div = div+1) begin
    @(negedge dclk);
    dclk_assert <= 1;
    for (i = 0; i < 5; i = i+1) @(negedge dclk);
    dclk_assert <= 0; 
  end
end
wire [3:0] switch;
assign switch = div-1;
  
// unit under test
wire [3:0] led;
wire clk;
wire rstN;
wire drstN;
wire compare;
TOP i_uut (
  .OSC_CLK     (osc_clk),
  .PO_RSTN     (po_rstN),
  .PB_RSTN     (pb_rstN),
  .VP          (vp),
  .VN          (vn),
  .SWITCH      (switch),
  .LED         (led),
  .LOCK        (lock),
  .CLK         (clk),
  .RSTN        (rstN),
  .DCLK        (dclk),
  .DRSTN       (drstN),
  .COMPARE     (compare));
  
// check period and duty cycle of CLK signal
real clk_rising;
real clk_falling;
real clk_half_period;
real clk_period;
always @(clk) begin
  if (clk == 1) begin
    clk_half_period = $realtime - clk_falling;
    clk_period      = $realtime - clk_rising;
    clk_rising      = $realtime;
  end else begin
    clk_half_period = $realtime - clk_rising;
    clk_period      = $realtime - clk_falling;
    clk_falling     = $realtime;
  end
  if (rstN == 1) begin
    testM(clk_half_period >= 0.45 * 250 && clk_half_period <= 0.55 * 250, "unexpected CLK duty cycle");
    testM(clk_period      >= 0.95 * 250 && clk_period      <= 1.05 * 250, "unexpected CLK period");
  end
end

// check number of CLK cycles
integer clk_cycles = 0;
always @(posedge clk) clk_cycles = clk_cycles + 1;
initial #199000 testM(clk_cycles > 790, "too few CLK cycles detected");


// check period and duty cycle of DCLK signal
real dclk_rising;
real dclk_falling;
real dclk_half_period;
real dclk_period;
always @(dclk) begin
  if (dclk == 1) begin
    dclk_half_period = $realtime - dclk_falling;
    dclk_period      = $realtime - dclk_rising;
    dclk_rising      = $realtime;
  end else begin
    dclk_half_period = $realtime - dclk_rising;
    dclk_period      = $realtime - dclk_falling;
    dclk_falling     = $realtime;
  end
  if (dclk_assert == 1) begin
    testM(dclk_half_period >= 0.32 * 250 * div && dclk_half_period <= 0.67*250*div, "unexpected DCLK duty cycle");
    testM(dclk_period      >= 0.95 * 250 * div && dclk_period      <= 1.05*250*div, "unexpected DCLK period");
  end
end

// check number of DCLK cycles
integer dclk_cycles = 0;
always @(posedge dclk) dclk_cycles = dclk_cycles + 1;
initial #199000 testM(dclk_cycles > 90, "too few DCLK cycles detected");

// check rstN signal
initial begin
  #30;  // IO delay (postlayout)
  testI(rstN, 0, "RSTN");
  @(posedge clk);
  #10;  // IO delay (postlayout)
  testI(rstN, 0, "RSTN");
  @(posedge clk);
  @(posedge clk);
  #10;  // IO delay (postlayout)
  testI(rstN, 1, "RSTN");
end

// check drstN signal
initial begin
  #35;  // IO delay (postlayout)
  testI(drstN, 0, "DRSTN");
  @(posedge dclk);
  #10;  // IO delay (postlayout)
  testI(drstN, 0, "DRSTN");
  @(posedge dclk);
  @(posedge dclk);
  #10;  // IO delay (postlayout)
  testI(drstN, 1, "DRSTN");
end

// check LED output
always @(led, switch) begin
  #19;  // IO delay (postlayout)
  testI(led, switch, "LED");
end  

// check LOCK signal
always @(po_rstN, pb_rstN, lock) begin
  #30;  // IO delay (postlayout)
  if (po_rstN & pb_rstN == 0) testI(lock, 0, "LOCK");
end
  
// check COMPARE signal (may only change after DCLK rising)
always @(compare) begin
  if (drstN == 1'b1) begin
    #1;
    testM($realtime - dclk_rising < 10, "COMPARE switched unexpected");
  end
end

endmodule 
