`default_nettype none
`timescale 1ns/100ps

// Simulation of the comparators switching behavior
// The LVDS and LVPECL input buffers can not be simulated. Thus, the OPAMP module generates combinatorical logic 
// (VP and not VN) instead of a differential input buffer. This module translates two real valued signals (the original
// comparator inputs to be compared against each other) into two logic signals to be passed to the OPAMP module.
// 
// author: Andreas Engel
// date:   03/13
module OPAMP_SIM #(`include "settings.v") 
 (input  wire [63:0] VPR,     // real valued positive comparator input        
  input  wire [63:0] VNR,     // real valued negative comparator input        
  output reg         VP,      // logic equivalent of positive comparator input
  output wire        VN);     // logic equivalent of negative comparator input
/**********************************************************************************************************************/  


real       diff_buffer;
real       diff = 0;
real       high;
real       low;
reg [17:0] lfsr = 0;

always begin
  diff         = ($bitstoreal(VPR) - $bitstoreal(VNR)) * 1000; // diff in mV
  diff_buffer <= diff;
  
  // edge detection for hysteresis simulation
  if (diff > diff_buffer) begin
    high = OPAMP_VRH;
    low  = OPAMP_VRL;
  end else begin
    high = OPAMP_VFH;
    low  = OPAMP_VFL;
  end
  
  // instable comparator decision inside switching limits
  if      (diff > high) VP <= 1;
  else if (diff < low)  VP <= 0;
  else                  VP <= lfsr[0];
  
  // pseudo random number generator
  lfsr <= {lfsr, lfsr[17] ~^ lfsr[10]};
  
  // fixed update rate
  #(OPAMP_UPDATE);
end

assign VN = 0;

endmodule