`default_nettype none
`timescale 1ns/100ps

// Testbench for system project.
//  * generates 48 MHz OSC_CLK and reset Signals (PO_RSTN, PB_RSTN)
//  * generates input voltage (1.15V, 1.35V)
//  * generate  gain settings (1, MAX_GAIN)
//  * simulate PGA behavior
//  * simulate comparator behavior
//  * simulate charge/discharge of Vcompare
//  * receive samples via UART
//  * calculate min / max / rms error of sampled values
// 
// autor: Andreas Engel
// date:  03/2013
module TB #(`include "settings.v");
`include "test.v"
/**********************************************************************************************************************/

// clock and reset generation
reg                   osc_clk;
initial               osc_clk = 0;
always  #(500.0/48.0) osc_clk = ~osc_clk;
reg                   po_rstN;
initial               po_rstN = 0;
initial #(300.0/48.0) po_rstN = 1;
reg                   pb_rstN;
initial               pb_rstN = 0;
initial #(200.0/48.0) pb_rstN = 1;
  

// RC-model
real vcompare = VCC;
wire charge;
always begin
  vcompare <= vcompare + ((charge ? VCC : 0) - vcompare) * (1.0 - $exp(- OPAMP_UPDATE / 1000.0 / TAU));
  #(OPAMP_UPDATE);
end

 
// simulate PGA behavior
wire [63:0] rscaled;
real        vscaled;
real        vin;
wire        sck;
wire        sdi;
wire        csN;
PGA_SIM i_pga (
  .SCK  (sck),
  .SDI  (sdi),
  .CSN  (csN),
  .VIN  ($realtobits(vin)),
  .VOUT (rscaled));

always @(rscaled) vscaled <= $bitstoreal(rscaled);

// check GAIN configuration of PGA
integer gain;
always @(gain) begin
  #3000; // SPI config time
  testR(rscaled, $realtobits(VREF + (vin-VREF)*((gain < MAX_GAIN) ? gain : MAX_GAIN)), "VSCALED");
end


// simulate comparator behavoir
wire compare;
wire scaled;
OPAMP_SIM i_opamp (
  .VPR ($realtobits(vcompare)),
  .VNR (rscaled),
  .VP  (compare),
  .VN  (scaled));


// combine user input to switch signal
reg        calibrate;
wire [4:0] switch = {calibrate,gain-1};
  
// unit under test
wire [4:0] led;
wire       rx;
TOP i_uut (
  .OSC_CLK  (osc_clk),
  .PB_RSTN  (pb_rstN),
  .PO_RSTN  (po_rstN),
  .SWITCH   (switch),
  .LED      (led),
  .TX       (rx),
  .VSCALED  (scaled),
  .VCOMPARE (compare),
  .VCHARGE  (charge),
  .SDI      (sdi),
  .SCK      (sck),
  .CSN      (csN));



// serial data reception
wire [7:0] byte ;
wire       valid;
URX i_byte (
  .RX    (rx),
  .VALID (valid),
  .BYTE  (byte));

// result calculation
localparam bytes   = (SAMPLE_WIDTH + 4 + 7) / 8;
integer    cnt     = 0;
integer    acc     = 0;
real       result  = 0;
integer    results = 0;
always @(posedge valid, posedge calibrate) begin
  if (calibrate)
    results <= 0;  
  else begin
    acc = acc + byte * 256**cnt;
    cnt = cnt + 1;
      
    if (cnt == bytes) begin // all bytes accumulated => calculate result
      result  <= VCC * acc / 2.0**(SAMPLE_WIDTH+4);
      results <= results + 1;
      acc      = 0;
      cnt      = 0;
    end 
  end 
end


// user interface: set input voltage, calibrate and gain
reg     end_of_sim = 0;
real    min;
real    max;
real    rms;
integer g,i;
initial begin
  calibrate  = 1'b0;
  vin        = 1.25+0.1;
  gain       = 1;
    
  // wait until PLL locked
  #10000;
    
  // sweep gain (in pseudo random order, to test MAX_GAIN saturation)
  for (g = 1; g <= 16; g = g+1) begin
    gain = g/2 + 8*(g % 2);
    #5000;
  end
  #5000;
    
  // check measurement accuracy for certain gains and input voltages
  for (g = 0; g <= 1; g = g+1) begin
    gain = 1+g*(MAX_GAIN-1);
    for (i = 0; i <= 1; i = i+1) begin
      vin <= 1.25 - 0.1 + i*0.2;
      @(results == 5);
      
      #10000;
      $display (" Vin = %f Gain = %f => rms error = %f %%", vin, gain, rms);
      
      calibrate = 1'b1;
      #4000;
      calibrate = 1'b0;
    end
  end
    
  // stop simulation
  #100000;
  $display("min/max error = %f %% / %f %%", min, max);
  end_of_sim = 1'b1;
end
    
//  calculate result error 
//+ local rms (for each GAIN/VIN combination) 
//+ global min/max error
real error;
real error_acc;
always @(results) begin
  if (results > 0) begin
    error = (result-vin) / vin * 100.0;
    testM(error*error <= 100, "Measurement error exceeds 10%");
    
    if (results == 1) error_acc = 0;
    error_acc = error_acc + error * error;
    rms <= $sqrt(error_acc/results);

    if (error > max) max <= error;
    if (error < min) min <= error;
  end
end

endmodule 
